"""
Integration tests for the Sphinx Notion Builder functionality.
"""

import json
import textwrap
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from beartype import beartype
from sphinx.testing.util import SphinxTestApp
from ultimate_notion import Emoji
from ultimate_notion.blocks import BulletedItem as UnoBulletedItem
from ultimate_notion.blocks import Callout as UnoCallout
from ultimate_notion.blocks import Code as UnoCode
from ultimate_notion.blocks import (
    Heading1 as UnoHeading1,
)
from ultimate_notion.blocks import (
    Heading2 as UnoHeading2,
)
from ultimate_notion.blocks import (
    Heading3 as UnoHeading3,
)
from ultimate_notion.blocks import Image as UnoImage
from ultimate_notion.blocks import (
    Paragraph as UnoParagraph,
)
from ultimate_notion.blocks import (
    Quote as UnoQuote,
)
from ultimate_notion.blocks import Table as UnoTable
from ultimate_notion.blocks import (
    TableOfContents as UnoTableOfContents,
)
from ultimate_notion.blocks import (
    ToggleItem as UnoToggleItem,
)
from ultimate_notion.core import NotionObject
from ultimate_notion.obj_api.core import GenericObject
from ultimate_notion.obj_api.enums import BGColor, CodeLang
from ultimate_notion.rich_text import text


@beartype
def _create_code_block_without_annotations(
    content: str, language: CodeLang
) -> UnoCode:
    """Create a code block without annotations to match the fixed behavior.

    This matches the fix in visit_literal_block where annotations are
    removed to prevent white text color in code blocks.
    """
    # Create rich text and remove annotations to match the fix
    code_text = text(text=content)
    # Remove annotations to prevent white text in code blocks
    del code_text.rich_texts[0].obj_ref.annotations  # pyright: ignore[reportUnknownMemberType]

    # Create code block and set the rich text
    code_block = UnoCode(text="dummy", language=language)
    code_block.rich_text = code_text
    return code_block


@beartype
def _reconstruct_nested_structure(
    items: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """
    Reconstruct the nested structure of the nodes from the graph produced.
    """
    result: list[dict[str, Any]] = []
    for item in items:
        block = item["block"]
        block_type = block["type"]

        # Handle blocks that can have children
        if block_type in {
            "paragraph",
            "callout",
            "bulleted_list_item",
            "toggle",
        }:
            children = item.get("children", [])
            nested_blocks = _reconstruct_nested_structure(
                items=children,
            )
            block[block_type]["children"] = nested_blocks

        result.append(block)
    return result


@beartype
def _assert_rst_converts_to_notion_objects(
    rst_content: str,
    expected_objects: list[NotionObject[Any]],
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
    extensions: tuple[str, ...] = ("sphinx_notion",),
) -> None:
    """
    The given rST content is converted to the given expected objects.
    """
    srcdir = tmp_path / "src"
    srcdir.mkdir()

    (srcdir / "conf.py").write_text(data="")

    cleaned_content = textwrap.dedent(text=rst_content).strip()
    (srcdir / "index.rst").write_text(data=cleaned_content)

    app = make_app(
        srcdir=srcdir,
        builddir=tmp_path / "build",
        buildername="notion",
        confoverrides={"extensions": list(extensions)},
    )
    app.build()

    output_file = app.outdir / "index.json"
    with output_file.open() as f:
        generated_json: list[dict[str, Any]] = json.load(fp=f)

    generated_json_un_flattened = _reconstruct_nested_structure(
        items=generated_json
    )

    expected_json: list[dict[str, Any]] = []
    for notion_object in expected_objects:
        obj_ref = notion_object.obj_ref
        assert isinstance(obj_ref, GenericObject)
        dumped_block = obj_ref.serialize_for_api()
        expected_json.append(dumped_block)

    assert generated_json_un_flattened == expected_json, (
        generated_json_un_flattened,
        expected_json,
    )


def test_single_paragraph(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Single paragraph converts to Notion JSON.
    """
    rst_content = """
        This is a simple paragraph for testing.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="This is a simple paragraph for testing.")
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_multiple_paragraphs(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Multiple paragraphs in convert to separate Notion blocks.
    """
    rst_content = """
        First paragraph with some text.

        Second paragraph with different content.

        Third paragraph to test multiple blocks.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="First paragraph with some text."),
        UnoParagraph(text="Second paragraph with different content."),
        UnoParagraph(text="Third paragraph to test multiple blocks."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_inline_formatting(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Inline formatting (bold, italic, code) converts to rich text.
    """
    rst_content = """
        This is **bold** and *italic* and ``inline code``.
    """

    normal_text = text(text="This is ")
    bold_text = text(text="bold", bold=True)
    normal_text2 = text(text=" and ")
    italic_text = text(text="italic", italic=True)
    normal_text3 = text(text=" and ")
    code_text = text(text="inline code", code=True)
    normal_text4 = text(text=".")

    combined_text = (
        normal_text
        + bold_text
        + normal_text2
        + italic_text
        + normal_text3
        + code_text
        + normal_text4
    )

    expected_paragraph = UnoParagraph(text="dummy")
    expected_paragraph.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [expected_paragraph]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_single_heading(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Single heading converts to Heading 1 block.
    """
    rst_content = """
        Main Title
        ==========

        This is content under the title.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoHeading1(text="Main Title"),
        UnoParagraph(text="This is content under the title."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_multiple_heading_levels(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Multiple heading levels convert to appropriate Notion heading blocks.
    """
    rst_content = """
        Main Title
        ==========

        Content under main title.

        Section Title
        -------------

        Content under section.

        Subsection Title
        ~~~~~~~~~~~~~~~~

        Content under subsection.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoHeading1(text="Main Title"),
        UnoParagraph(text="Content under main title."),
        UnoHeading2(text="Section Title"),
        UnoParagraph(text="Content under section."),
        UnoHeading3(text="Subsection Title"),
        UnoParagraph(text="Content under subsection."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_heading_with_formatting(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Headings with inline formatting convert to rich text in heading blocks.
    """
    rst_content = """
        **Bold** and *Italic* Title
        ============================

        Content follows.
    """

    bold_text = text(text="Bold", bold=True)
    normal_text = text(text=" and ")
    italic_text = text(text="Italic", italic=True)
    normal_text2 = text(text=" Title")

    combined_text = bold_text + normal_text + italic_text + normal_text2

    expected_heading = UnoHeading1(text="dummy")
    expected_heading.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [
        expected_heading,
        UnoParagraph(text="Content follows."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_simple_link(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Simple links convert to rich text with href.
    """
    rst_content = """
        This paragraph contains a `link to example <https://example.com>`_.
    """

    normal_text1 = text(text="This paragraph contains a ")
    link_text = text(text="link to example", href="https://example.com")
    normal_text2 = text(text=".")

    combined_text = normal_text1 + link_text + normal_text2

    expected_paragraph = UnoParagraph(text="dummy")
    expected_paragraph.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [expected_paragraph]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_multiple_links(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Multiple links in a paragraph convert correctly.
    """
    # Write proper rST content to file to avoid Python string escaping issues
    rst_file = tmp_path / "test_content.rst"
    content = (
        "Visit `Google <https://google.com>`_ and "
        "`GitHub <https://github.com>`_\ntoday."
    )
    rst_file.write_text(data=content)
    rst_content = rst_file.read_text()

    normal_text1 = text(text="Visit ")
    link_text1 = text(text="Google", href="https://google.com")
    normal_text2 = text(text=" and ")
    link_text2 = text(text="GitHub", href="https://github.com")
    normal_text3 = text(text="\ntoday.")

    combined_text = (
        normal_text1 + link_text1 + normal_text2 + link_text2 + normal_text3
    )

    expected_paragraph = UnoParagraph(text="dummy")
    expected_paragraph.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [expected_paragraph]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_link_in_heading(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Links in headings convert to rich text with href.
    """
    rst_content = """
        Check out `Notion API <https://developers.notion.com>`_
        ========================================================

        Content follows.
    """

    normal_text1 = text(text="Check out ")
    link_text = text(text="Notion API", href="https://developers.notion.com")

    combined_text = normal_text1 + link_text

    expected_heading = UnoHeading1(text="dummy")
    expected_heading.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [
        expected_heading,
        UnoParagraph(text="Content follows."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_mixed_formatting_with_links(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Links mixed with other formatting work correctly.
    """
    rst_content = """
        This has **bold** and a `link <https://example.com>`_ and *italic*.
    """

    normal_text1 = text(text="This has ")
    bold_text = text(text="bold", bold=True)
    normal_text2 = text(text=" and a ")
    link_text = text(text="link", href="https://example.com")
    normal_text3 = text(text=" and ")
    italic_text = text(text="italic", italic=True)
    normal_text4 = text(text=".")

    combined_text = (
        normal_text1
        + bold_text
        + normal_text2
        + link_text
        + normal_text3
        + italic_text
        + normal_text4
    )

    expected_paragraph = UnoParagraph(text="dummy")
    expected_paragraph.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [expected_paragraph]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_unnamed_link_with_backticks(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Unnamed links with backticks convert to rich text with href.

    The text should be just the URL without angle brackets.
    """
    rst_content = """
        Visit `<https://example.com>`_ for more information.
    """

    normal_text1 = text(text="Visit ")
    link_text = text(text="https://example.com", href="https://example.com")
    normal_text2 = text(text=" for more information.")

    combined_text = normal_text1 + link_text + normal_text2

    expected_paragraph = UnoParagraph(text="dummy")
    expected_paragraph.rich_text = combined_text

    expected_objects: list[NotionObject[Any]] = [expected_paragraph]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_simple_quote(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that block quotes convert to Notion Quote blocks.
    """
    rst_content = """
        Regular paragraph.

            This is a block quote.

        Another paragraph.
    """
    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        UnoQuote(text="This is a block quote."),
        UnoParagraph(text="Another paragraph."),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_multiline_quote(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that multi-line block quotes convert to Notion Quote blocks.
    """
    rst_content = """
        Regular paragraph.

            This is a multi-line
            block quote with
            multiple lines.

        Another paragraph.
    """
    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        UnoQuote(
            text="This is a multi-line\nblock quote with\nmultiple lines."
        ),
        UnoParagraph(text="Another paragraph."),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_table_of_contents(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that contents directive converts to Notion TableOfContents blocks.
    """
    rst_content = """
        Introduction
        ============

        .. contents::

        First Section
        -------------

        Content here.

        Second Section
        --------------

        More content.
    """
    expected_objects: list[NotionObject[Any]] = [
        UnoHeading1(text="Introduction"),
        UnoTableOfContents(),
        UnoHeading2(text="First Section"),
        UnoParagraph(text="Content here."),
        UnoHeading2(text="Second Section"),
        UnoParagraph(text="More content."),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_toctree_directive(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that the ``toctree`` directive converts to nothing.
    """
    rst_content = """
        Introduction
        ============

        .. toctree::
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoHeading1(text="Introduction"),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_simple_code_block(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Code blocks convert to Notion Code blocks.
    """
    rst_content = """
        Regular paragraph.

        .. code-block:: python

           def hello():
               print("Hello, world!")

        Another paragraph.
    """
    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        _create_code_block_without_annotations(
            content='def hello():\n    print("Hello, world!")',
            language=CodeLang.PYTHON,
        ),
        UnoParagraph(text="Another paragraph."),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_code_block_language_mapping(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that various languages map correctly.
    """
    rst_content = """
        .. code-block:: console

           $ pip install example

        .. code-block:: javascript

           console.log("hello");

        .. code-block:: bash

           echo "test"

        .. code-block:: text

           Some plain text

        .. code-block::

           Code with no language
    """
    expected_objects: list[NotionObject[Any]] = [
        _create_code_block_without_annotations(
            content="$ pip install example", language=CodeLang.SHELL
        ),
        _create_code_block_without_annotations(
            content='console.log("hello");', language=CodeLang.JAVASCRIPT
        ),
        _create_code_block_without_annotations(
            content='echo "test"', language=CodeLang.BASH
        ),
        _create_code_block_without_annotations(
            content="Some plain text", language=CodeLang.PLAIN_TEXT
        ),
        _create_code_block_without_annotations(
            content="Code with no language", language=CodeLang.PLAIN_TEXT
        ),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_flat_bullet_list(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that flat bullet lists convert correctly to Notion BulletedItem.
    """
    rst_content = """
        * First bullet point
        * Second bullet point
        * Third bullet point with longer text
    """
    expected_objects: list[NotionObject[Any]] = [
        UnoBulletedItem(text="First bullet point"),
        UnoBulletedItem(text="Second bullet point"),
        UnoBulletedItem(text="Third bullet point with longer text"),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_bullet_list_with_inline_formatting(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test bullet lists with inline formatting.
    """
    rst_content = """
        * This is **bold text** in a bullet
    """
    bullet = UnoBulletedItem(text="placeholder")
    bullet.rich_text = (
        text(text="This is ", bold=False, italic=False, code=False)
        + text(text="bold text", bold=True, italic=False, code=False)
        + text(text=" in a bullet", bold=False, italic=False, code=False)
    )

    expected_objects: list[NotionObject[Any]] = [
        bullet,
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    argnames=("admonition_type", "emoji", "background_color", "message"),
    argvalues=[
        ("note", "📝", BGColor.BLUE, "This is an important note."),
        ("warning", "⚠️", BGColor.YELLOW, "This is a warning message."),
        ("tip", "💡", BGColor.GREEN, "This is a helpful tip."),
    ],
)
def test_admonition_single_line(
    *,
    admonition_type: str,
    emoji: str,
    background_color: BGColor,
    message: str,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that admonitions convert to Notion Callout blocks.
    """
    rst_content = f"""
        Regular paragraph.

        .. {admonition_type}:: {message}

        Another paragraph.
    """

    callout = UnoCallout(
        text=message,
        icon=Emoji(emoji=emoji),
        color=background_color,
    )

    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        callout,
        UnoParagraph(text="Another paragraph."),
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    argnames=("admonition_type", "emoji", "background_color"),
    argvalues=[
        ("note", "📝", BGColor.BLUE),
        ("warning", "⚠️", BGColor.YELLOW),
        ("tip", "💡", BGColor.GREEN),
    ],
)
def test_admonition_multiline(
    admonition_type: str,
    emoji: str,
    background_color: BGColor,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Test that admonitions with multiple paragraphs work.

    The first paragraph becomes the callout text, and subsequent
    paragraphs become nested blocks within the callout.
    """
    rst_content = f"""
        .. {admonition_type}::
           This is the first paragraph of the {admonition_type}.

           This is the second paragraph that should be nested.
    """
    callout = UnoCallout(
        text="",
        icon=Emoji(emoji=emoji),
        color=background_color,
    )
    callout.rich_text = text(
        text=f"This is the first paragraph of the {admonition_type}."
    )

    nested_paragraph = UnoParagraph(
        text="This is the second paragraph that should be nested."
    )

    callout.obj_ref.value.children.append(nested_paragraph.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    expected_objects: list[NotionObject[Any]] = [
        callout,
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_admonition_with_code_block(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that admonitions can contain code blocks as nested children.
    """
    rst_content = """
        .. note::
           This note contains a code example.

           .. code-block:: python

              def hello():
                  print("Hello, world!")

           The code above demonstrates a simple function.
    """

    callout = UnoCallout(text="", icon=Emoji(emoji="📝"), color=BGColor.BLUE)
    callout.rich_text = text(text="This note contains a code example.")

    nested_code_block = _create_code_block_without_annotations(
        content='def hello():\n    print("Hello, world!")',
        language=CodeLang.PYTHON,
    )
    nested_paragraph = UnoParagraph(
        text="The code above demonstrates a simple function."
    )

    callout.obj_ref.value.children.append(nested_code_block.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    callout.obj_ref.value.children.append(nested_paragraph.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    expected_objects: list[NotionObject[Any]] = [
        callout,
    ]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_admonition_with_code_block_first(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """Test admonition with code block as first child (not paragraph).

    This tests the else clause when first child is not a paragraph.
    """
    rst_content = """
        .. note::

           .. code-block:: python

              def hello():
                  print("Hello, world!")

           This paragraph comes after the code block.
    """

    callout = UnoCallout(text="", icon=Emoji(emoji="📝"), color=BGColor.BLUE)
    callout.rich_text = text(text="")

    nested_code_block = _create_code_block_without_annotations(
        content='def hello():\n    print("Hello, world!")',
        language=CodeLang.PYTHON,
    )
    nested_paragraph = UnoParagraph(
        text="This paragraph comes after the code block."
    )

    callout.obj_ref.value.children.append(nested_code_block.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    callout.obj_ref.value.children.append(nested_paragraph.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    expected_objects: list[NotionObject[Any]] = [callout]
    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_admonition_with_bullet_points(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that bullet points show up within admonitions (issue #78).
    """
    rst_content = """
        .. note::

           This is an important note that demonstrates the note admonition
           support.

           * A
           * B
    """

    callout = UnoCallout(text="", icon=Emoji(emoji="📝"), color=BGColor.BLUE)
    callout.rich_text = text(
        text="This is an important note that demonstrates the note "
        "admonition\nsupport."
    )

    bullet_a = UnoBulletedItem(text="A")
    bullet_b = UnoBulletedItem(text="B")

    callout.obj_ref.value.children.append(bullet_a.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    callout.obj_ref.value.children.append(bullet_b.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    expected_objects: list[NotionObject[Any]] = [
        callout,
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_nested_bullet_list(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that deeply nested bullet lists work.
    """
    rst_content = """
        * Top level item
        * Top level with children

          * Second level item
          * Second level with children

            * Third level item (now allowed!)

        * Another top level item
    """

    third_level_1 = UnoBulletedItem(text="Third level item (now allowed!)")

    second_level_1 = UnoBulletedItem(text="Second level item")
    second_level_2 = UnoBulletedItem(text="Second level with children")

    top_level_1 = UnoBulletedItem(text="Top level item")
    top_level_2 = UnoBulletedItem(text="Top level with children")

    # Remove pyright ignore once we have
    # https://github.com/ultimate-notion/ultimate-notion/issues/94.
    second_level_2.obj_ref.value.children.append(third_level_1.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    top_level_2.obj_ref.value.children.append(second_level_1.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    top_level_2.obj_ref.value.children.append(second_level_2.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    top_level_3 = UnoBulletedItem(text="Another top level item")

    expected_objects: list[NotionObject[Any]] = [
        top_level_1,
        top_level_2,
        top_level_3,
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
        extensions=("sphinx_notion", "sphinx_toolbox.collapse"),
    )


def test_collapse_block(
    *,
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that collapse directives convert to Notion ToggleItem blocks.
    """
    rst_content = """
        Regular paragraph.

        .. collapse:: Click to expand

           This content is hidden by default.

           It supports **formatting**.

        Another paragraph.
    """

    toggle_block = UnoToggleItem(text="Click to expand")

    nested_para1 = UnoParagraph(text="This content is hidden by default.")
    nested_para2 = UnoParagraph(text="It supports formatting.")
    nested_para2.rich_text = (
        text(text="It supports ", bold=False)
        + text(text="formatting", bold=True)
        + text(text=".", bold=False)
    )

    toggle_block.obj_ref.value.children.append(nested_para1.obj_ref)  # pyright: ignore[reportUnknownMemberType]
    toggle_block.obj_ref.value.children.append(nested_para2.obj_ref)  # pyright: ignore[reportUnknownMemberType]

    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        toggle_block,
        UnoParagraph(text="Another paragraph."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
        extensions=("sphinx_notion", "sphinx_toolbox.collapse"),
    )


def test_simple_table(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Simple rST table converts to Notion Table block.
    """
    rst_content = """
        +----------+----------+
        | Header 1 | Header 2 |
        +==========+==========+
        | Cell 1   | Cell 2   |
        +----------+----------+
        | Cell 3   | Cell 4   |
        |          |          |
        | Cell 3   | Cell 4   |
        +----------+----------+
    """

    table = UnoTable(n_rows=3, n_cols=2, header_row=True)
    # Header row
    table[0, 0] = text(text="Header 1")
    table[0, 1] = text(text="Header 2")
    # First data row
    table[1, 0] = text(text="Cell 1")
    table[1, 1] = text(text="Cell 2")
    # Second data row
    table[2, 0] = text(text="Cell 3\n\nCell 3")
    table[2, 1] = text(text="Cell 4\n\nCell 4")

    expected_objects: list[NotionObject[Any]] = [table]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_table_without_header_row(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Table without a heading row converts to Notion Table block with
    header_row=False.
    """
    rst_content = """
        +--------+--------+
        | Cell 1 | Cell 2 |
        +--------+--------+
        | Cell 3 | Cell 4 |
        +--------+--------+
    """
    table = UnoTable(n_rows=2, n_cols=2, header_row=False)
    table[0, 0] = text(text="Cell 1")
    table[0, 1] = text(text="Cell 2")
    table[1, 0] = text(text="Cell 3")
    table[1, 1] = text(text="Cell 4")

    expected_objects: list[NotionObject[Any]] = [table]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_table_inline_formatting(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Table headers and cells with inline formatting (bold, italic, code) convert
    to Notion Table with rich text in header and cells.
    """
    rst_content = """
        +----------------------+----------------------+
        | **Header Bold**      | *Header Italic*      |
        +======================+======================+
        | ``cell code``        | Normal cell          |
        +----------------------+----------------------+
    """

    table = UnoTable(n_rows=2, n_cols=2, header_row=True)

    table[0, 0] = text(text="Header Bold", bold=True)
    table[0, 1] = text(text="Header Italic", italic=True)

    table[1, 0] = text(text="cell code", code=True)
    table[1, 1] = text(text="Normal cell")

    expected_objects: list[NotionObject[Any]] = [table]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_simple_image(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that image directives convert to Notion Image blocks.
    """
    rst_content = """
        Regular paragraph.

        .. image:: https://www.example.com/path/to/image.png

        Another paragraph.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoParagraph(text="Regular paragraph."),
        UnoImage(url="https://www.example.com/path/to/image.png"),
        UnoParagraph(text="Another paragraph."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )


def test_image_with_alt_text_only(
    make_app: Callable[..., SphinxTestApp],
    tmp_path: Path,
) -> None:
    """
    Test that image directives with only alt text convert to Notion Image
    blocks without captions.
    """
    rst_content = """
        .. image:: https://www.example.com/path/to/image.png
           :alt: Example image

        Content follows.
    """

    expected_objects: list[NotionObject[Any]] = [
        UnoImage(url="https://www.example.com/path/to/image.png"),
        UnoParagraph(text="Content follows."),
    ]

    _assert_rst_converts_to_notion_objects(
        rst_content=rst_content,
        expected_objects=expected_objects,
        make_app=make_app,
        tmp_path=tmp_path,
    )



"""
Sphinx Notion Builder.
"""

import json
from functools import singledispatchmethod
from typing import Any

from beartype import beartype
from docutils import nodes
from docutils.nodes import NodeVisitor
from sphinx.application import Sphinx
from sphinx.builders.text import TextBuilder
from sphinx.util.typing import ExtensionMetadata
from sphinx_toolbox.collapse import CollapseNode
from ultimate_notion import Emoji
from ultimate_notion.blocks import BulletedItem as UnoBulletedItem
from ultimate_notion.blocks import Callout as UnoCallout
from ultimate_notion.blocks import Code as UnoCode
from ultimate_notion.blocks import Heading as UnoHeading
from ultimate_notion.blocks import (
    Heading1 as UnoHeading1,
)
from ultimate_notion.blocks import (
    Heading2 as UnoHeading2,
)
from ultimate_notion.blocks import (
    Heading3 as UnoHeading3,
)
from ultimate_notion.blocks import Image as UnoImage
from ultimate_notion.blocks import (
    Paragraph as UnoParagraph,
)
from ultimate_notion.blocks import (
    Quote as UnoQuote,
)
from ultimate_notion.blocks import Table as UnoTable
from ultimate_notion.blocks import (
    TableOfContents as UnoTableOfContents,
)
from ultimate_notion.blocks import (
    ToggleItem as UnoToggleItem,
)
from ultimate_notion.core import NotionObject
from ultimate_notion.obj_api.core import GenericObject
from ultimate_notion.obj_api.enums import BGColor, CodeLang
from ultimate_notion.rich_text import Text, text

type _BlockTree = dict[tuple[NotionObject[Any], int], "_BlockTree"]


@beartype
def _create_rich_text_from_children(*, node: nodes.Element) -> Text:
    """Create Notion rich text from ``docutils`` node children.

    This uses ``ultimate-notion``'s rich text capabilities to
    avoid some size limits.

    See: https://developers.notion.com/reference/request-limits#size-limits.
    """
    rich_text = Text.from_plain_text(text="")

    for child in node.children:
        if isinstance(child, nodes.reference):
            link_url = child.attributes["refuri"]
            link_text = child.attributes.get("name", link_url)

            new_text = text(
                text=link_text,
                href=link_url,
                bold=False,
                italic=False,
                code=False,
            )
        elif isinstance(child, nodes.target):
            continue
        else:
            new_text = text(
                text=child.astext(),
                bold=isinstance(child, nodes.strong),
                italic=isinstance(child, nodes.emphasis),
                code=isinstance(child, nodes.literal),
            )
        rich_text += new_text

    return rich_text


@beartype
def _extract_table_structure(
    node: nodes.table,
) -> tuple[int, nodes.row | None, list[nodes.row]]:
    """
    Return (n_cols, header_row, body_rows) for a table node.
    """
    header_row = None
    body_rows: list[nodes.row] = []
    n_cols = 0

    for child in node.children:
        assert isinstance(child, nodes.tgroup)
        n_cols = int(child.get(key="cols", failobj=0))
        for tgroup_child in child.children:
            if isinstance(tgroup_child, nodes.thead):
                for row in tgroup_child.children:
                    assert isinstance(row, nodes.row)
                    header_row = row
            elif isinstance(tgroup_child, nodes.tbody):
                for row in tgroup_child.children:
                    assert isinstance(row, nodes.row)
                    body_rows.append(row)

    return n_cols, header_row, body_rows


@beartype
def _cell_source_node(entry: nodes.Node) -> nodes.paragraph:
    """Return the paragraph child of an entry if present, else the entry.

    This isolates the small branch used when converting a table cell so
    the main table function becomes simpler.
    """
    paragraph_children = [
        c for c in entry.children if isinstance(c, nodes.paragraph)
    ]
    if len(paragraph_children) == 1:
        return paragraph_children[0]

    # If there are multiple children (multiple paragraphs or mixed nodes),
    # create a combined node that preserves all content. We insert a
    # double-newline text node between each top-level child to mimic
    # paragraph separation when converting to plain text/rich text.
    # Join the plain text of each top-level child with two newlines so
    # the resulting rich text becomes a single text fragment like
    # 'Cell 3\n\nCell 3' (matches test expectations).
    parts: list[str] = [c.astext() for c in entry.children]
    joined = "\n\n".join(parts)
    combined = nodes.paragraph()
    combined += nodes.Text(data=joined)
    return combined


@beartype
def _map_pygments_to_notion_language(*, pygments_lang: str) -> CodeLang:
    """
    Map ``Pygments`` language names to Notion CodeLang ``enum`` values.
    """
    language_mapping: dict[str, CodeLang] = {
        "abap": CodeLang.ABAP,
        "arduino": CodeLang.ARDUINO,
        "bash": CodeLang.BASH,
        "basic": CodeLang.BASIC,
        "c": CodeLang.C,
        "clojure": CodeLang.CLOJURE,
        "coffeescript": CodeLang.COFFEESCRIPT,
        "console": CodeLang.SHELL,
        "cpp": CodeLang.CPP,
        "c++": CodeLang.CPP,
        "csharp": CodeLang.CSHARP,
        "c#": CodeLang.CSHARP,
        "css": CodeLang.CSS,
        "dart": CodeLang.DART,
        "default": CodeLang.PLAIN_TEXT,
        "diff": CodeLang.DIFF,
        "docker": CodeLang.DOCKER,
        "dockerfile": CodeLang.DOCKER,
        "elixir": CodeLang.ELIXIR,
        "elm": CodeLang.ELM,
        "erlang": CodeLang.ERLANG,
        "flow": CodeLang.FLOW,
        "fortran": CodeLang.FORTRAN,
        "fsharp": CodeLang.FSHARP,
        "f#": CodeLang.FSHARP,
        "gherkin": CodeLang.GHERKIN,
        "glsl": CodeLang.GLSL,
        "go": CodeLang.GO,
        "graphql": CodeLang.GRAPHQL,
        "groovy": CodeLang.GROOVY,
        "haskell": CodeLang.HASKELL,
        "html": CodeLang.HTML,
        "java": CodeLang.JAVA,
        "javascript": CodeLang.JAVASCRIPT,
        "js": CodeLang.JAVASCRIPT,
        "json": CodeLang.JSON,
        "julia": CodeLang.JULIA,
        "kotlin": CodeLang.KOTLIN,
        "latex": CodeLang.LATEX,
        "tex": CodeLang.LATEX,
        "less": CodeLang.LESS,
        "lisp": CodeLang.LISP,
        "livescript": CodeLang.LIVESCRIPT,
        "lua": CodeLang.LUA,
        "makefile": CodeLang.MAKEFILE,
        "make": CodeLang.MAKEFILE,
        "markdown": CodeLang.MARKDOWN,
        "md": CodeLang.MARKDOWN,
        "markup": CodeLang.MARKUP,
        "matlab": CodeLang.MATLAB,
        "mermaid": CodeLang.MERMAID,
        "nix": CodeLang.NIX,
        "objective-c": CodeLang.OBJECTIVE_C,
        "objc": CodeLang.OBJECTIVE_C,
        "ocaml": CodeLang.OCAML,
        "pascal": CodeLang.PASCAL,
        "perl": CodeLang.PERL,
        "php": CodeLang.PHP,
        "powershell": CodeLang.POWERSHELL,
        "ps1": CodeLang.POWERSHELL,
        "prolog": CodeLang.PROLOG,
        "protobuf": CodeLang.PROTOBUF,
        "python": CodeLang.PYTHON,
        "py": CodeLang.PYTHON,
        "r": CodeLang.R,
        "reason": CodeLang.REASON,
        "ruby": CodeLang.RUBY,
        "rb": CodeLang.RUBY,
        "rust": CodeLang.RUST,
        "rs": CodeLang.RUST,
        "sass": CodeLang.SASS,
        "scala": CodeLang.SCALA,
        "scheme": CodeLang.SCHEME,
        "scss": CodeLang.SCSS,
        "shell": CodeLang.SHELL,
        "sh": CodeLang.SHELL,
        "sql": CodeLang.SQL,
        "swift": CodeLang.SWIFT,
        "text": CodeLang.PLAIN_TEXT,
        "toml": CodeLang.TOML,
        "typescript": CodeLang.TYPESCRIPT,
        "ts": CodeLang.TYPESCRIPT,
        "vb.net": CodeLang.VB_NET,
        "vbnet": CodeLang.VB_NET,
        "verilog": CodeLang.VERILOG,
        "vhdl": CodeLang.VHDL,
        "visual basic": CodeLang.VISUAL_BASIC,
        "vb": CodeLang.VISUAL_BASIC,
        "webassembly": CodeLang.WEBASSEMBLY,
        "wasm": CodeLang.WEBASSEMBLY,
        "xml": CodeLang.XML,
        "yaml": CodeLang.YAML,
        "yml": CodeLang.YAML,
    }

    return language_mapping[pygments_lang.lower()]


@beartype
class NotionTranslator(NodeVisitor):
    """
    Translate ``docutils`` nodes to Notion JSON.
    """

    def __init__(self, document: nodes.document, builder: TextBuilder) -> None:
        """
        Initialize the translator with storage for blocks.
        """
        del builder
        super().__init__(document=document)
        self._block_tree: _BlockTree = {}
        self.body: str
        self._section_level = 0

    @beartype
    def _add_block_to_tree(
        self,
        block: NotionObject[Any],
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """Add a block to the block tree.

        First has to find the parent in the tree recursively.
        """
        block_key = (block, id(block))
        if not parent_path:
            self._block_tree[block_key] = {}
            return

        current_node = self._block_tree[parent_path[0]]

        for parent_key in parent_path[1:]:
            current_node = current_node[parent_key]
        current_node[block_key] = {}

    @beartype
    def _process_list_item_recursively(
        self,
        *,
        node: nodes.list_item,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Recursively process a list item node and return a BulletedItem.
        """
        paragraph = node.children[0]
        assert isinstance(paragraph, nodes.paragraph)
        rich_text = _create_rich_text_from_children(node=paragraph)
        block = UnoBulletedItem(text="placeholder")
        block.rich_text = rich_text
        self._add_block_to_tree(
            block=block,
            parent_path=parent_path,
        )

        bullet_only_msg = (
            "The only thing Notion supports within a bullet list is a "
            "bullet list"
        )
        assert isinstance(node, nodes.list_item)

        for child in node.children[1:]:
            assert isinstance(child, nodes.bullet_list), bullet_only_msg
            for nested_list_item in child.children:
                assert isinstance(nested_list_item, nodes.list_item), (
                    bullet_only_msg
                )
                self._process_list_item_recursively(
                    node=nested_list_item,
                    parent_path=[*parent_path, (block, id(block))],
                )

    @singledispatchmethod
    @beartype
    def _process_node_to_blocks(  # pylint: disable=no-self-use
        self,
        node: nodes.Element,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:  # pragma: no cover
        """
        Required function for ``singledispatch``.
        """
        del section_level
        del parent_path
        raise NotImplementedError(node)

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.table,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """Process rST table nodes by creating Notion Table blocks.

        This implementation delegates small branches to helpers which
        keeps the function body linear and easier to reason about.
        """
        del section_level

        n_cols, header_row, body_rows = _extract_table_structure(node=node)

        n_rows = 1 + len(body_rows) if header_row else len(body_rows)
        table = UnoTable(
            n_rows=n_rows, n_cols=n_cols, header_row=bool(header_row)
        )

        row_idx = 0
        if header_row is not None:
            for col_idx, entry in enumerate(iterable=header_row.children):
                source = _cell_source_node(entry=entry)
                table[row_idx, col_idx] = _create_rich_text_from_children(
                    node=source
                )
            row_idx += 1

        for body_row in body_rows:
            for col_idx, entry in enumerate(iterable=body_row.children):
                source = _cell_source_node(entry=entry)
                table[row_idx, col_idx] = _create_rich_text_from_children(
                    node=source
                )
            row_idx += 1

        self._add_block_to_tree(block=table, parent_path=parent_path)

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.paragraph,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process paragraph nodes by creating Notion Paragraph blocks.
        """
        del section_level
        rich_text = _create_rich_text_from_children(node=node)
        paragraph_block = UnoParagraph(text="")
        paragraph_block.rich_text = rich_text
        self._add_block_to_tree(block=paragraph_block, parent_path=parent_path)

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.block_quote,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process block quote nodes by creating Notion Quote blocks.
        """
        del section_level
        rich_text = _create_rich_text_from_children(node=node)
        quote_block = UnoQuote(text="")
        quote_block.rich_text = rich_text
        self._add_block_to_tree(block=quote_block, parent_path=parent_path)

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.literal_block,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process literal block nodes by creating Notion Code blocks.
        """
        del section_level
        code_text = _create_rich_text_from_children(node=node)
        pygments_lang = node.get(key="language", failobj="")
        language = _map_pygments_to_notion_language(
            pygments_lang=pygments_lang,
        )
        code_block = UnoCode(text=code_text, language=language)
        # By default, the code block has a color set (DEFAULT) which means
        # that there is no syntax highlighting.
        # See https://github.com/ultimate-notion/ultimate-notion/issues/93.
        for rich_text in code_text.rich_texts:  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
            del rich_text.obj_ref.annotations  # pyright: ignore[reportUnknownMemberType]
        code_block.rich_text = code_text
        self._add_block_to_tree(block=code_block, parent_path=parent_path)

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.bullet_list,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process bullet list nodes by creating Notion BulletedItem blocks.
        """
        del section_level
        bullet_only_msg = (
            "The only thing Notion supports within a bullet list is a "
            "bullet list"
        )
        for list_item in node.children:
            assert isinstance(list_item, nodes.list_item), bullet_only_msg
            self._process_list_item_recursively(
                node=list_item,
                parent_path=parent_path,
            )

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.topic,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process topic nodes, specifically for table of contents.
        """
        del section_level  # Not used for topics
        # Later, we can support `.. topic::` directives, likely as
        # a callout with no icon.
        assert "contents" in node["classes"]
        toc_block = UnoTableOfContents()
        self._add_block_to_tree(block=toc_block, parent_path=parent_path)

    @_process_node_to_blocks.register
    def _(  # pylint: disable=no-self-use
        self,
        node: nodes.compound,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process Sphinx ``toctree`` nodes.
        """
        del node
        del section_level
        del parent_path
        # There are no specific Notion blocks for ``toctree`` nodes.
        # We need to support ``toctree`` in ``index.rst``.
        # Just ignore it.

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.title,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process title nodes by creating appropriate Notion heading blocks.
        """
        rich_text = _create_rich_text_from_children(node=node)

        heading_levels: dict[int, type[UnoHeading[Any]]] = {
            1: UnoHeading1,
            2: UnoHeading2,
            3: UnoHeading3,
        }
        heading_cls = heading_levels[section_level]
        block = heading_cls(text="")

        block.rich_text = rich_text
        self._add_block_to_tree(block=block, parent_path=parent_path)

    @beartype
    def _create_admonition_callout(
        self,
        node: nodes.Element,
        *,
        emoji: str,
        background_color: BGColor,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """Create a Notion Callout block for admonition nodes.

        The first child (typically a paragraph) becomes the callout
        text, and any remaining children become nested blocks within the
        callout.
        """
        block = UnoCallout(
            text="", icon=Emoji(emoji=emoji), color=background_color
        )

        self._add_block_to_tree(block=block, parent_path=parent_path)
        # Use the first child as the callout text
        first_child = node.children[0]
        if isinstance(first_child, nodes.paragraph):
            rich_text = _create_rich_text_from_children(node=first_child)
            block.rich_text = rich_text
            # Process remaining children as nested blocks
            children_to_process = node.children[1:]
        else:
            # If first child is not a paragraph, use empty text
            block.rich_text = Text.from_plain_text(text="")
            # Process all children as nested blocks (including the first)
            children_to_process = node.children

        # Process children as nested blocks
        for child in children_to_process:
            self._process_node_to_blocks(
                child,
                section_level=1,
                parent_path=[*parent_path, (block, id(block))],
            )

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.note,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process note admonition nodes by creating Notion Callout blocks.
        """
        del section_level
        self._create_admonition_callout(
            node=node,
            emoji="📝",
            background_color=BGColor.BLUE,
            parent_path=parent_path,
        )

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.warning,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process warning admonition nodes by creating Notion Callout blocks.
        """
        del section_level
        self._create_admonition_callout(
            node=node,
            emoji="⚠️",
            background_color=BGColor.YELLOW,
            parent_path=parent_path,
        )

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.tip,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process tip admonition nodes by creating Notion Callout blocks.
        """
        del section_level
        self._create_admonition_callout(
            node=node,
            emoji="💡",
            background_color=BGColor.GREEN,
            parent_path=parent_path,
        )

    @_process_node_to_blocks.register
    def _(
        self,
        node: CollapseNode,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process collapse nodes by creating Notion ToggleItem blocks.
        """
        del section_level

        title_text = node.attributes["label"]
        toggle_block = UnoToggleItem(text=title_text)
        self._add_block_to_tree(block=toggle_block, parent_path=parent_path)

        for child in node.children:
            self._process_node_to_blocks(
                child,
                section_level=1,
                parent_path=[*parent_path, (toggle_block, id(toggle_block))],
            )

    @_process_node_to_blocks.register
    def _(
        self,
        node: nodes.image,
        *,
        section_level: int,
        parent_path: list[tuple[NotionObject[Any], int]],
    ) -> None:
        """
        Process image nodes by creating Notion Image blocks.
        """
        del section_level

        image_url = node.attributes["uri"]

        # Create the image block (alt text is not used as caption)
        image_block = UnoImage(url=image_url, caption=None)
        self._add_block_to_tree(block=image_block, parent_path=parent_path)

    def visit_title(self, node: nodes.Element) -> None:
        """
        Handle title nodes by creating appropriate Notion heading blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_section(self, node: nodes.Element) -> None:
        """
        Handle section nodes by increasing the section level.
        """
        del node
        self._section_level += 1

    def depart_section(self, node: nodes.Element) -> None:
        """
        Handle leaving section nodes by decreasing the section level.
        """
        del node
        self._section_level -= 1

    def visit_paragraph(self, node: nodes.Element) -> None:
        """
        Handle paragraph nodes by creating Notion Paragraph blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_block_quote(self, node: nodes.Element) -> None:
        """
        Handle block quote nodes by creating Notion Quote blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_literal_block(self, node: nodes.Element) -> None:
        """
        Handle literal block nodes by creating Notion Code blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_bullet_list(self, node: nodes.Element) -> None:
        """
        Handle bullet list nodes by processing each list item.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_topic(self, node: nodes.Element) -> None:
        """
        Handle topic nodes, specifically for table of contents.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_note(self, node: nodes.Element) -> None:
        """
        Handle note admonition nodes by creating Notion Callout blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_warning(self, node: nodes.Element) -> None:
        """
        Handle warning admonition nodes by creating Notion Callout blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_tip(self, node: nodes.Element) -> None:
        """
        Handle tip admonition nodes by creating Notion Callout blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_table(self, node: nodes.Element) -> None:
        """
        Handle table nodes by creating Notion Table blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_compound(self, node: nodes.Element) -> None:
        """
        Handle compound admonition nodes by creating a table of contents block.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_CollapseNode(self, node: nodes.Element) -> None:  # pylint: disable=invalid-name  # noqa: N802
        """
        Handle collapse nodes by creating Notion ToggleItem blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_image(self, node: nodes.Element) -> None:
        """
        Handle image nodes by creating Notion Image blocks.
        """
        self._process_node_to_blocks(
            node,
            section_level=self._section_level,
            parent_path=[],
        )

        raise nodes.SkipNode

    def visit_document(self, node: nodes.Element) -> None:
        """
        Initialize block collection at document start.
        """
        assert self
        del node

    @beartype
    def _convert_block_tree_to_json(
        self,
        # `beartype` does not support recursive types, so we need to use a
        # simpler type.
        block_tree: dict[tuple[NotionObject[Any], int], Any],
    ) -> list[dict[str, Any]]:
        """
        Convert the block tree to a JSON-serializable format, ignoring IDs from
        tuples.
        """
        result: list[dict[str, Any]] = []
        for (block, _), subtree in block_tree.items():
            obj_ref = block.obj_ref
            assert isinstance(obj_ref, GenericObject)
            dumped_structure = {
                "block": obj_ref.serialize_for_api(),
                "children": self._convert_block_tree_to_json(
                    block_tree=subtree
                ),
            }
            result.append(dumped_structure)
        return result

    def depart_document(self, node: nodes.Element) -> None:
        """
        Output collected block tree as JSON at document end.
        """
        del node

        json_output = json.dumps(
            obj=self._convert_block_tree_to_json(block_tree=self._block_tree),
            indent=2,
            ensure_ascii=False,
        )
        self.body = json_output


@beartype
class NotionBuilder(TextBuilder):
    """
    Build Notion-compatible documents.
    """

    name = "notion"
    out_suffix = ".json"


@beartype
def setup(app: Sphinx) -> ExtensionMetadata:
    """
    Add the builder to Sphinx.
    """
    app.add_builder(builder=NotionBuilder)
    app.set_translator(name="notion", translator_class=NotionTranslator)

    return {"parallel_read_safe": True}



"""Upload documentation to Notion.

Inspired by https://github.com/ftnext/sphinx-notion/blob/main/upload.py.
"""

import argparse
import json
import os
import sys
from pathlib import Path
from typing import Any

from beartype import beartype
from notion_client import Client
from ultimate_notion import Session
from ultimate_notion.page import Page

_NOTION_BLOCKS_BATCH_SIZE = 100  # Max blocks per request to avoid 413 errors


_Block = dict[str, Any]


@beartype
def _find_existing_page_by_title(parent_page: Page, title: str) -> Page | None:
    """Find an existing page with the given title in the parent page (top-level
    only).

    Return the page if found, otherwise None.
    """
    for child_page in parent_page.subpages:
        if str(object=child_page.title) == title:
            return child_page
    return None


@beartype
def _upload_blocks_in_batches(
    notion_client: Client,
    parent_id: str,
    blocks: list[_Block],
    batch_size: int,
) -> None:
    """
    Upload blocks to a page in batches to avoid 413 errors.
    """
    total_blocks = len(blocks)
    sys.stderr.write(
        f"Uploading {total_blocks} blocks in batches of {batch_size}...\n"
    )

    for i in range(0, total_blocks, batch_size):
        batch = blocks[i : i + batch_size]
        batch_num = (i // batch_size) + 1
        total_batches = (total_blocks + batch_size - 1) // batch_size

        sys.stderr.write(
            f"Uploading batch {batch_num}/{total_batches} "
            f"({len(batch)} blocks)...\n"
        )

        notion_client.blocks.children.append(
            block_id=parent_id,
            children=batch,
        )

    sys.stderr.write(f"Successfully uploaded all {total_blocks} blocks.\n")


@beartype
def _parse_args() -> argparse.Namespace:
    """
    Parse command-line arguments for the upload script.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Upload to Notion",
    )
    parser.add_argument(
        "-f",
        "--file",
        help="JSON File to upload",
        required=True,
        type=Path,
    )
    parser.add_argument(
        "-p",
        "--parent_page_id",
        help="Parent page ID (integration connected)",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--title",
        help="Title of the new page",
        required=True,
    )
    parser.add_argument(
        "--batch-size",
        help="Number of blocks per batch",
        type=int,
        default=_NOTION_BLOCKS_BATCH_SIZE,
    )
    return parser.parse_args()


@beartype
def upload_blocks_recursively(
    parent_id: str,
    block_details_list: list[dict[str, Any]],
    notion_client: Client,
    batch_size: int,
) -> None:
    """
    Upload blocks recursively, handling the new structure with block and
    children.
    """
    if not block_details_list:
        return

    # Extract just the blocks for this level
    level_blocks = [details["block"] for details in block_details_list]

    # Upload this level's blocks in batches
    _upload_blocks_in_batches(
        notion_client=notion_client,
        parent_id=parent_id,
        blocks=level_blocks,
        batch_size=batch_size,
    )

    # Get the uploaded blocks to get their IDs for children
    uploaded_blocks: dict[str, Any] = notion_client.blocks.children.list(  # type: ignore[assignment] # pyright: ignore[reportAssignmentType]
        block_id=parent_id
    )
    block_id_map: dict[str, list[dict[str, Any]]] = {}

    # Map the uploaded blocks to their details for children processing
    results = uploaded_blocks.get("results", [])
    for i, block in enumerate(iterable=results):
        if i < len(block_details_list):
            block_details = block_details_list[i]
            if block_details["children"]:
                block_id = block.get("id")
                if block_id:
                    block_id_map[str(object=block_id)] = block_details[
                        "children"
                    ]

    # Recursively upload children for each block that has them
    for block_id, children in block_id_map.items():
        upload_blocks_recursively(
            parent_id=block_id,
            block_details_list=children,
            notion_client=notion_client,
            batch_size=batch_size,
        )


@beartype
def main() -> None:
    """
    Main entry point for the upload command.
    """
    args = _parse_args()

    notion_client = Client(auth=os.environ["NOTION_TOKEN"])
    session = Session(client=notion_client)
    batch_size = args.batch_size
    title = args.title
    file_path = args.file

    blocks = json.loads(s=file_path.read_text(encoding="utf-8"))

    # Workaround for https://github.com/ultimate-notion/ultimate-notion/issues/103
    Page.parent_db = None  # type: ignore[method-assign,assignment] # pyright: ignore[reportAttributeAccessIssue]
    assert Page.parent_db is None

    parent_page = session.get_page(page_ref=args.parent_page_id)
    page = _find_existing_page_by_title(
        parent_page=parent_page, title=args.title
    )

    if not page:
        page = session.create_page(parent=parent_page, title=args.title)
        sys.stdout.write(f"Created new page: {args.title} (ID: {page.id})\n")

    for child in list(  # pyright: ignore[reportUnknownVariableType]
        page.children  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
    ):
        child.delete()

    # Start the recursive upload process
    upload_blocks_recursively(
        parent_id=str(object=page.id),
        block_details_list=blocks,
        notion_client=notion_client,
        batch_size=batch_size,
    )
    sys.stdout.write(f"Updated existing page: {title} (ID: {page.id})\n")


if __name__ == "__main__":
    main()
